/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <obs/obs_stack.h>
#include <string.h>

OBS_DEF_INLINE_FUNC_(OBS_size, calc_blocks, (OBS_stack *st, OBS_size size))
{
	OBS_size blocksize = obs_stack_get_block_size(st), newsize;
	OBS_bool dbl = (st->features & OBS_STACK_F_EXPONENTIAL_EXPAND) != 0;

	if (dbl) {
		newsize = blocksize;
		while (newsize && newsize < size)
			newsize <<= 1;
		if (newsize == 0)
			return 0;
	} else {
		if (size == 0)
			newsize = blocksize;
		else {
			newsize = size + blocksize - 1;
			if (newsize < size)
				return 0;
		}
	}
	return newsize / blocksize;
}

static OBS_status
expand(OBS_stack *st, OBS_size desired_size, OBS_bool *isfrag)
{
	OBS_size blocks;
	OBS_region newreg, lastreg;
	OBS_status ret;
	OBS_stack_chunk *ch, *prev;

	if (isfrag)
		*isfrag = OBS_FALSE;

	blocks = calc_blocks(st, desired_size + sizeof (OBS_stack_chunk));
	if (blocks == 0)
		return OBS_S_TOO_BIG;

	if (st->last) {
		prev = st->last->prev;
		obs_stack_chunk_to_block_(st->last, &newreg);
		if ((st->features & OBS_STACK_F_NO_CONCATENATE) == 0) {
			ret = obs_pa_expand_tail(
				&st->pagealloc, &newreg, blocks);
			if (ret == OBS_S_NOERR)
				goto expanded;
			else if (ret != OBS_S_NOTSUPP &&
				 ret != OBS_S_CANNOT_EXPAND)
				return ret;
		}
	}

	ret = obs_pa_alloc(&st->pagealloc, &newreg, blocks);
	if (ret)
		return ret;
	if (st->last && (st->features & OBS_STACK_F_NO_CONCATENATE) == 0) {
		obs_stack_chunk_to_block_(st->last, &lastreg);
		ret = obs_pa_concatenate(&st->pagealloc, &lastreg, &newreg);
		if (ret == OBS_S_NOT_CONTIGUOUS ||
		    ret == OBS_S_NOTSUPP) {
			if (isfrag)
				*isfrag = OBS_TRUE;
		} else if (ret) {
			obs_pa_free(&st->pagealloc, &newreg);
			return ret;
		} else /* concatenation succeeded */
			newreg = lastreg;
	} else if (isfrag)
		*isfrag = OBS_TRUE;
	prev = st->last;

expanded:
	ch = obs_stack_block_to_chunk_(&newreg);
	ch->size = obs_stack_block_to_chunk_size_(&newreg);
	ch->prev = prev;
	st->last = ch;

	return OBS_S_NOERR;
}

#define AWARE_PA_FEATURES (OBS_PA_F_EXPAND_TAIL |			      \
			   OBS_PA_F_SHRINK_HEAD |			      \
			   OBS_PA_F_SHRINK_TAIL |			      \
			   OBS_PA_F_ALL_FREE_ON_UNINIT |		      \
			   OBS_PA_F_CONCATENATE)

OBS_status
obs_stack_init_pa_ops(OBS_stack *st, OBS_size numunit, OBS_flags features,
		      int align, const OBS_pa_ops *paops)
{
	OBS_status ret;
	OBS_flags pafeat;
	int maxalign;

	pafeat = obs_pa_ops_get_available_features(paops);
	pafeat &= AWARE_PA_FEATURES;
	maxalign = obs_pa_ops_get_alignment(paops);
	if (align<1 || align > maxalign)
		return OBS_S_INVAL;

	ret = obs_pa_ops_init(paops, &st->pagealloc, numunit, pafeat);
	if (ret)
		return ret;
	st->alignment = align;
	st->alignment_mask = (1 << align) - 1;
	st->last = NULL;
	st->features = features;
	obs_region_init(&st->top);

	return OBS_S_NOERR;
}

void
obs_stack_uninit(OBS_stack *st)
{
	if ((obs_pa_get_features(&st->pagealloc) &
	     OBS_PA_F_ALL_FREE_ON_UNINIT) == 0)
		obs_stack_free_all(st);
	obs_pa_uninit(&st->pagealloc);
	memset(st, 0, sizeof (OBS_stack));
}

OBS_status
obs_stack_truncate_after(OBS_stack *st, void *ptr)
{
	OBS_stack_chunk *curlast = st->last;

	while (st->last != NULL &&
	       (ptr == NULL ||
		!obs_stack_is_pointer_in_chunk(st->last, ptr))) {
		OBS_stack_chunk *prev = st->last->prev;
		OBS_region reg;
		obs_stack_chunk_to_block_(st->last, &reg);
		obs_pa_free(&st->pagealloc, &reg);
		st->last = prev;
	}
	if (st->last == NULL) {
		obs_region_init(&st->top);
		if (ptr)
			return OBS_S_REVOKED;
		return OBS_S_NOERR;
	}

	if (st->last != curlast || (char *)st->top.head > (char *)ptr) {
		ptr = (void *)obs_stack_round_up_(st, (OBS_uintptr)ptr);
		st->top.head = ptr;
		st->top.size = 0;
	} else if (obs_region_is_valid(&st->top) &&
		   (char *)ptr - (char *)st->top.head < st->top.size)
		st->top.size = (char *)ptr - (char *)st->top.head;

	if ((st->features & OBS_STACK_F_NO_SHRINK_AFTER) == 0 &&
	    (obs_pa_get_features(&st->pagealloc) & OBS_PA_F_SHRINK_TAIL)!= 0) {
		OBS_size blocksize = obs_stack_get_block_size(st);
		OBS_size desired_size =
		    (char *)ptr-(char *)obs_stack_get_chunk_begin_(st->last) +
		    sizeof (OBS_stack_chunk);
		OBS_size newsize = calc_blocks(st, desired_size) * blocksize;
		OBS_region reg;

		obs_stack_chunk_to_block_(st->last, &reg);
		if (newsize < reg.size) {
			OBS_stack_chunk *prevch = st->last->prev;
			OBS_status ret;

			ret = obs_pa_shrink_tail(&st->pagealloc, &reg,
						 (reg.size-newsize)/blocksize);
			if (ret == OBS_S_NOERR) {
				OBS_stack_chunk *newch =
				    obs_stack_block_to_chunk_(&reg);
				newch->size =
				    obs_stack_block_to_chunk_size_(&reg);
				newch->prev = prevch;
				st->last = newch;
			}
		}
	}

	return OBS_S_NOERR;
}

OBS_status
obs_stack_truncate_before(OBS_stack *st, void *ptr)
{
	OBS_stack_chunk *ch, *prev, *curch;

	for (curch=st->last;
	     curch && !obs_stack_is_pointer_in_chunk(curch, ptr);
	     curch=curch->prev)
		;

	if (!curch)
		return ptr ? OBS_S_INVAL : OBS_S_NOERR;

	prev = curch->prev;
	curch->prev = NULL;
	while ((ch = prev) != NULL) {
		OBS_region reg;
		prev = ch->prev;
		obs_stack_chunk_to_block_(ch, &reg);
		obs_pa_free(&st->pagealloc, &reg);
	}

	ptr = (void *)obs_stack_round_down_(st, (OBS_uintptr)ptr);
	if (st->last == curch) {
		if (!obs_region_is_valid(&st->top) ||
		    (char *)st->top.head + st->top.size < (char *)ptr) {
			ptr = obs_stack_get_chunk_end_(curch);
			obs_region_init(&st->top);
		} else if ((char *)st->top.head < (char *)ptr) {
			st->top.size -= (char *)ptr - (char *)st->top.head;
			st->top.head = ptr;
		}
	}

	if ((st->features & OBS_STACK_F_NO_SHRINK_BEFORE) == 0 &&
	    (obs_pa_get_features(&st->pagealloc) & OBS_PA_F_SHRINK_HEAD)!= 0) {
		OBS_size blocksize = obs_stack_get_block_size(st);
		OBS_size desired_size =
		    (char *)obs_stack_get_chunk_end_(curch) - (char *)ptr +
		    sizeof (OBS_stack_chunk);
		OBS_size newsize = calc_blocks(st, desired_size) * blocksize;
		OBS_region reg;

		obs_stack_chunk_to_block_(curch, &reg);
		if (newsize < reg.size) {
			OBS_status ret;

			ret = obs_pa_shrink_head(&st->pagealloc, &reg,
						 (reg.size-newsize)/blocksize);
			if (ret == OBS_S_NOERR) {
				curch->size =
				    obs_stack_block_to_chunk_size_(&reg);
			}
		}
	}
	if (!obs_region_is_valid(&st->top))
		st->top.head = obs_stack_get_chunk_begin_(st->last);

	return OBS_S_NOERR;
}

OBS_status
obs_stack_set_alignment(OBS_stack *st, int align)
{
	int maxalign;

	maxalign = obs_stack_get_max_alignment(st);
	if (align<1 || align > maxalign)
		return OBS_S_INVAL;

	if (st->top.head)
		st->top.head = (void *)((char *)st->top.head + st->top.size);
	st->top.size = 0;

	return OBS_S_NOERR;
}

OBS_status
obs_stack_alloc_(OBS_stack *st, OBS_size size)
{
	OBS_size desired_size = size;
	OBS_bool isfrag;
	OBS_status ret;

	if (obs_region_is_valid(&st->top))
		desired_size +=
		    obs_stack_get_aligned_next_(st) -
		    (OBS_uintptr)obs_stack_get_chunk_begin_(st->last);

	ret = expand(st, desired_size, &isfrag);
	if (ret)
		return ret;

	if (isfrag)
		st->top.head = obs_stack_get_chunk_begin_(st->last);
	else
		st->top.head = (void *)obs_stack_get_aligned_next_(st);
	st->top.size = size;

	return OBS_S_NOERR;
}

static void
handle_orphaned_chunk(OBS_stack *st)
{
	OBS_status ret;
	OBS_region ophreg, curreg;
	OBS_stack_chunk *orphaned = st->last->prev;
	OBS_stack_chunk *prev = orphaned->prev;

	/* ASSERT: st->last->prev == orphaned */
	obs_stack_chunk_to_block_(st->last, &curreg);
	obs_stack_chunk_to_block_(orphaned, &ophreg);

	ret = OBS_S_NOT_CONTIGUOUS;
	if ((st->features & OBS_STACK_F_NO_CONCATENATE) == 0)
		ret = obs_pa_concatenate(&st->pagealloc, &curreg, &ophreg);
	if (ret == OBS_S_NOERR) {
		/* merge */
		OBS_stack_chunk *ch;
		ch = obs_stack_block_to_chunk_(&curreg);
		ch->size = obs_stack_block_to_chunk_size_(&curreg);
		st->last = ch;
	} else /* free */
		obs_pa_free(&st->pagealloc, &ophreg);
	st->last->prev = prev;
}

static void
shrink_reloc(OBS_stack *st)
{
	OBS_stack_chunk *leaved;
	OBS_size desired_size, blocksize, newsize;
	OBS_region reg;

	if (st->top.head == NULL ||
	    (st->features & OBS_STACK_F_NO_SHRINK_RELOC) != 0 ||
	    (obs_pa_get_features(&st->pagealloc) & OBS_PA_F_SHRINK_TAIL) == 0)
		return;

	leaved = st->last->prev;
	desired_size =
	    (char *)st->top.head - (char *)obs_stack_get_chunk_begin_(leaved) +
	    sizeof (OBS_stack_chunk);
	blocksize = obs_stack_get_block_size(st);
	newsize = desired_size + blocksize - 1;
	newsize -= newsize % blocksize;
	obs_stack_chunk_to_block_(leaved, &reg);
	if (newsize < reg.size) {
		OBS_stack_chunk *prevch = leaved->prev;
		OBS_status ret;

		ret = obs_pa_shrink_tail(&st->pagealloc, &reg,
					 (reg.size-newsize)/blocksize);
		if (ret == OBS_S_NOERR) {
			leaved = obs_stack_block_to_chunk_(&reg);
			leaved->size = obs_stack_block_to_chunk_size_(&reg);
			leaved->prev = prevch;
			st->last->prev = leaved;
		}
	}
}


OBS_status
obs_stack_reserve_top_(OBS_stack *st, OBS_size desired_size)
{
	OBS_bool isfrag;
	OBS_status ret;
	OBS_bool issole = obs_stack_is_sole_in_last_chunk_(st);

	if (st->last && st->top.head)
		desired_size +=
		    (OBS_uintptr)st->top.head -
		    (OBS_uintptr)obs_stack_get_chunk_begin_(st->last);

	ret = expand(st, desired_size, &isfrag);
	if (ret)
		return ret;
	if (isfrag) {
		void *newtop = obs_stack_get_chunk_begin_(st->last);
		if (st->last->prev) {
			if (st->top.head)
				memcpy(newtop, st->top.head, st->top.size);
			if (issole)
				handle_orphaned_chunk(st);
			else
				shrink_reloc(st);
		}
		st->top.head = newtop;
	}

	return OBS_S_NOERR;
}
