/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef OBS_STACK_H_
#define OBS_STACK_H_

#include <obs/obs_pagealloc.h>
#include <string.h>

#define OBS_STACK_F_EXPONENTIAL_EXPAND	0x0001
#define OBS_STACK_F_NO_CONCATENATE	0x0002
#define OBS_STACK_F_NO_SHRINK_AFTER	0x0004
#define OBS_STACK_F_NO_SHRINK_BEFORE	0x0008
#define OBS_STACK_F_NO_SHRINK_RELOC	0x0010

typedef struct OBS_stack OBS_stack;
typedef struct OBS_stack_chunk OBS_stack_chunk;

struct OBS_stack_chunk
{
	OBS_stack_chunk *prev;
	OBS_size size;
};

struct OBS_stack
{
	OBS_pa pagealloc;
	OBS_stack_chunk *last;
	OBS_flags features;
	OBS_region top;
	int alignment;
	OBS_uintptr alignment_mask;
};

OBS_BEGIN_DECL_
OBS_status obs_stack_init_pa_ops(OBS_stack *, OBS_size, OBS_flags,
				 int, const OBS_pa_ops *);
void obs_stack_uninit(OBS_stack *);
OBS_status obs_stack_truncate_after(OBS_stack *, void *);
OBS_status obs_stack_truncate_before(OBS_stack *, void *);
OBS_status obs_stack_set_alignment(OBS_stack *, int);
OBS_status obs_stack_reserve_top_(OBS_stack *, OBS_size);
OBS_END_DECL_

OBS_DEF_INLINE_FUNC_(void, obs_stack_free_all, (OBS_stack *st))
{ obs_stack_truncate_after(st, NULL); }

OBS_DEF_INLINE_FUNC_(void, obs_stack_revoke_top, (OBS_stack *st))
{ st->top.size = 0; }

OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_init,
		     (OBS_stack *st, OBS_size blocks, OBS_flags features,
		      int alignment))
{
	return obs_stack_init_pa_ops(
		st, blocks, features, alignment,
		obs_pa_get_current_ops());
}

OBS_DEF_INLINE_FUNC_(OBS_flags, obs_stack_get_features,
		     (const OBS_stack *st))
{ return st->features; }

OBS_DEF_INLINE_FUNC_(const OBS_region *, obs_stack_get_top,
		     (const OBS_stack *st))
{ return &st->top; }

OBS_DEF_INLINE_FUNC_(OBS_size, obs_stack_get_block_size,
		     (const OBS_stack *st))
{ return obs_pa_get_block_size(&st->pagealloc); }

OBS_DEF_INLINE_FUNC_(int, obs_stack_get_max_alignment, (const OBS_stack *st))
{ return obs_pa_ops_get_alignment(obs_pa_get_ops(&st->pagealloc)); }

OBS_DEF_INLINE_FUNC_(int, obs_stack_get_alignment, (const OBS_stack *st))
{ return st->alignment; }

OBS_DEF_INLINE_FUNC_(void *, obs_stack_get_chunk_begin_, (OBS_stack_chunk *ch))
{
	return (void *)((char *)(void *)ch - ch->size);
}

OBS_DEF_INLINE_FUNC_(void *, obs_stack_get_chunk_end_, (OBS_stack_chunk *ch))
{
	return (void *)ch;
}

OBS_DEF_INLINE_FUNC_(void, obs_stack_chunk_to_block_,
		     (OBS_stack_chunk *ch, OBS_region *reg))
{
	reg->head = obs_stack_get_chunk_begin_(ch);
	reg->size = ch->size + sizeof (OBS_stack_chunk);
}

OBS_DEF_INLINE_FUNC_(OBS_stack_chunk *, obs_stack_block_to_chunk_,
		     (const OBS_region *reg))
{
	return obs_region_right(reg, sizeof (OBS_stack_chunk));
}

OBS_DEF_INLINE_FUNC_(void, obs_stack_chunk_to_data_,
		     (OBS_stack_chunk *ch, OBS_region *reg))
{
	reg->head = obs_stack_get_chunk_begin_(ch);
	reg->size = ch->size;
}

OBS_DEF_INLINE_FUNC_(OBS_size, obs_stack_block_to_chunk_size_,
		     (const OBS_region *reg))
{
	return reg->size - sizeof (OBS_stack_chunk);
}

OBS_DEF_INLINE_FUNC_(OBS_bool, obs_stack_is_pointer_in_chunk,
		     (OBS_stack_chunk *ch, void *ptr))
{
	return
	    (char *)obs_stack_get_chunk_begin_(ch) <= (char *)ptr &&
	    (char *)ptr < (char *)obs_stack_get_chunk_end_(ch);
}

OBS_DEF_INLINE_FUNC_(OBS_uintptr, obs_stack_round_down_,
		     (const OBS_stack *st, OBS_uintptr ptr))
{
	return ptr & ~st->alignment_mask;
}

OBS_DEF_INLINE_FUNC_(OBS_uintptr, obs_stack_round_up_,
		     (const OBS_stack *st, OBS_uintptr ptr))
{
	return (ptr + st->alignment_mask) & ~st->alignment_mask;
}

OBS_DEF_INLINE_FUNC_(OBS_uintptr, obs_stack_get_aligned_next_,
		     (OBS_stack *st))
{
	return obs_stack_round_up_(
		st, (OBS_uintptr)st->top.head + st->top.size);
}

OBS_DEF_INLINE_FUNC_(OBS_bool, obs_stack_is_sole_in_last_chunk_,
		     (OBS_stack *st))
{
	return
	    st->last &&
	    st->top.head == obs_stack_get_chunk_begin_(st->last);
}


OBS_DEF_INLINE_FUNC_(OBS_size, obs_stack_get_free_size, (OBS_stack *st))
{
	if (!st->last)
		return 0;
	else if (!obs_region_is_valid(&st->top))
		return st->last->size;

	return
	    (char *)obs_stack_get_chunk_end_(st->last) -
	    (char *)obs_region_end(&st->top);
}

OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_reserve_top,
		     (OBS_stack *st, OBS_size size))
{
	OBS_uintptr p;

	if (!st->last)
		return obs_stack_reserve_top_(st, size);

	if (st->top.size >= size)
		return OBS_S_NOERR;

	if (!st->top.head)
		st->top.head = obs_stack_get_chunk_begin_(st->last);
	else if (st->top.size == 0)
		st->top.head = (void *)obs_stack_get_aligned_next_(st);

	p = (OBS_uintptr)st->top.head;
	if (p + size < p)
		return OBS_S_INVAL; /* overflow */
	if (p + size > (OBS_uintptr)obs_stack_get_chunk_end_(st->last))
		return obs_stack_reserve_top_(st, size);

	return OBS_S_NOERR;
}

OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_alloc,
		     (OBS_stack *st, OBS_size size))
{
	OBS_status ret;

	if (obs_region_is_valid(&st->top)) {
		st->top.head = (void *)obs_stack_get_aligned_next_(st);
		st->top.size = 0;
	}
	ret = obs_stack_reserve_top(st, size);
	if (ret)
		return ret;
	st->top.size = size;

	return OBS_S_NOERR;
}

OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_resize_top,
		     (OBS_stack *st, OBS_size size))
{
	OBS_status ret;

	if (st->top.size == 0)
		return obs_stack_alloc(st, size);

	ret = obs_stack_reserve_top(st, size);
	if (ret)
		return ret;

	st->top.size = size;

	return OBS_S_NOERR;
}

OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_expand_top,
		     (OBS_stack *st, OBS_size incr))
{
	if (st->top.size == 0)
		return obs_stack_alloc(st, incr);
	if (st->top.size + incr < st->top.size)
		return OBS_S_INVAL;
	return obs_stack_resize_top(st, st->top.size + incr);
}

OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_append_top,
		     (OBS_stack *st, const void *data, OBS_size size))
{
	OBS_status ret = obs_stack_expand_top(st, size);
	if (ret)
		return ret;
	memcpy(obs_region_right(&st->top, size), data, size);
	return OBS_S_NOERR;
}

OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_append_1_top,
		     (OBS_stack *st, char ch))
{
	OBS_status ret = obs_stack_expand_top(st, 1);
	if (ret)
		return ret;
	*(char *)obs_region_right(&st->top, 1) = ch;
	return OBS_S_NOERR;
}

OBS_DEF_INLINE_FUNC_(void, obs_stack_resize_top_fast,
		     (OBS_stack *st, OBS_size size))
{ st->top.size = size; }

OBS_DEF_INLINE_FUNC_(void, obs_stack_expand_top_fast,
		     (OBS_stack *st, OBS_size incr))
{ st->top.size += incr; }

OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_append_top_fast,
		     (OBS_stack *st, const void *data, OBS_size size))
{
	memcpy(obs_region_left(&st->top, st->top.size), data, size);
	st->top.size += size;
	return OBS_S_NOERR;
}

OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_append_1_top_fast,
		     (OBS_stack *st, char ch))
{
	*(char *)obs_region_left(&st->top, st->top.size) = ch;
	st->top.size++;
	return OBS_S_NOERR;
}

#endif
